/****************************************************************************

 AMTransmitter
		Copyright(C) Denshi Usagi

****************************************************************************/

// PIC16F18346 Configuration Bit Settings

// CONFIG1
#pragma config FEXTOSC = OFF    // FEXTOSC External Oscillator mode Selection bits (Oscillator not enabled)
#pragma config RSTOSC = HFINT1  // Power-up default value for COSC bits (HFINTOSC (1MHz))
#pragma config CLKOUTEN = OFF   // Clock Out Enable bit (CLKOUT function is disabled; I/O or oscillator function on OSC2)
#pragma config CSWEN = ON       // Clock Switch Enable bit (Writing to NOSC and NDIV is allowed)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config MCLRE = ON       // Master Clear Enable bit (MCLR/VPP pin function is MCLR; Weak pull-up enabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable bits (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config LPBOREN = OFF    // Low-power BOR enable bit (ULPBOR disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable bits (Brown-out Reset enabled, SBOREN bit ignored)
#pragma config BORV = LOW       // Brown-out Reset Voltage selection bit (Brown-out voltage (Vbor) set to 2.45V)
#pragma config PPS1WAY = ON     // PPSLOCK bit One-Way Set Enable bit (The PPSLOCK bit can be cleared and set only once; PPS registers remain locked after one clear/set cycle)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable bit (Stack Overflow or Underflow will cause a Reset)
#pragma config DEBUG = OFF      // Debugger enable bit (Background debugger disabled)

// CONFIG3
#pragma config WRT = OFF        // User NVM self-write protection bits (Write protection off)
#pragma config LVP = ON         // Low Voltage Programming Enable bit (Low Voltage programming enabled. MCLR/VPP pin function is MCLR. MCLRE configuration bit is ignored.)

// CONFIG4
#pragma config CP = OFF         // User NVM Program Memory Code Protection bit (User NVM code protection disabled)
#pragma config CPD = OFF        // Data NVM Memory Code Protection bit (Data NVM code protection disabled)

#include <xc.h>
#include "typedefs.h"
#include "ddstbl.h"

//=================================================================

#define ADDR_LCD		0b0111110

#define AD9833_SYNC1	LATC5
#define AD9833_SYNC2	LATC6
#define AD9833_SCLK		LATC4
#define AD9833_SDAT		LATC3

#define LED				LATC7
#define SW1				RB7

#define BAT_V			0b000101 // ANA5
#define AGC_V			0b001101 // ANB5

#define RE1_A			RC0
#define RE1_B			RA2
#define RE2_A			RC2
#define RE2_B			RC1

#define RE_NONE			0
#define RE1_UP			1
#define RE1_DOWN		2
#define RE2_UP			3
#define RE2_DOWN		4

#define ADDR_FRQ		0
#define ADDR_SIG		1
#define DDS_FRQ			0
#define DDS_SIG			1

#define LOBAT			700UL	// バッテリー終止電圧ADC値(4.5V)
#define SIGDISPTIME		(65535 - 62000) // 約2秒

//=================================================================

typedef union {
    struct {
		BYTE			:8;
		BYTE			:4;
		BYTE D12		:1;
		BYTE D13		:1;
		BYTE D14		:1;
		BYTE D15		:1;
	};
    struct {
		BYTE 			:1;
		BYTE MODE		:1;
		BYTE 			:1;
		BYTE DIV2		:1;
		BYTE 			:1;
		BYTE OPBITEN	:1;
		BYTE SLEEP12	:1;
		BYTE SLEEP1		:1;
		BYTE RESET		:1;
		BYTE 			:1;
		BYTE PSEL		:1;
		BYTE FSEL		:1;
		BYTE HLB		:1;
		BYTE B28		:1;
		BYTE			:1;
		BYTE			:1;
	};
	WORD wValue;
} AD9833reg_t;

static AD9833reg_t AD9833[2];
static int g_nFrq;
static int g_nSig;
static WORD g_wMaxLv;

//=================================================================

static void InitPeripheral(void);
static void InitSystem(void);
static void Run(void);

static void Print(BYTE y, BYTE r, WORD wVal);

static int GetRE(void);
static int GetRE1(void);
static int GetRE2(void);
static BOOL GetSW(void);
static WORD GetADC(BYTE nCh);

static void SetLED(int nVal);

static void AD9833_Reset(BYTE nNo);
static void AD9833_SetFreq(BYTE nNo, DWORD dwVal);
static void AD9833_Start(BYTE nNo);
static void AD9833_Send(BYTE nNo, WORD wData);

static void InitLCD(void);
static void SetLCD(BYTE y, BYTE x, const char* pDat);
static void CmdLCD(BYTE bCmd);
static void I2CStart(void);
static void I2CSend(BYTE bData);
static void I2CStop(void);

static void WriteEEP(BYTE nAdd, BYTE nData);
static BYTE ReadEEP(BYTE nAdd);

static void DelayMs(WORD wMs);
#define DelayNs(x) // FOSCは8MHzなのでNSオーダーの待ちは不可

//=================================================================

// ●メインエントリ
void main(void)
{
	InitPeripheral();
	InitSystem();
	Run();
}

// ●ペリフェラル初期化
static void InitPeripheral(void)
{
	// Clock(8MHz)
	OSCFRQbits.HFFRQ = 0b0100;
	OSCCON1bits.NDIV = 0;
	
	// ポート初期化 (All Digital input)
	ANSELA = 0; ANSELB = 0; ANSELC = 0;
	LATA = 0; LATB = 0; LATC = 0;
	
	// AD9833
	AD9833_SYNC1 = 1;
	AD9833_SYNC2 = 1;
	TRISCbits.TRISC6 = 0;
	TRISCbits.TRISC3 = 0;
	TRISCbits.TRISC4 = 0;
	TRISCbits.TRISC5 = 0;

	// SW1
	WPUBbits.WPUB7 = 1;
	
	// LED
	TRISCbits.TRISC7 = 0;
	
	// ADC
	ANSELAbits.ANSA5 = 1; // BAT-V
	ANSELBbits.ANSB5 = 1; // AGC-V
	
	ADCON1bits.ADPREF = 0; // VREF+ is connected to VDD
	ADCON1bits.ADNREF = 0; // VREF- is connected to AVSS
	ADCON1bits.ADCS = 0b111; // dedicated RC oscillator
	ADCON1bits.ADFM = 1; // Right justified. 
	ADCON0bits.ADON = 1;

	// Timer2(Delay)
	T2CONbits.T2OUTPS = 0b1001; // 1:10 Postscaler
	PR2 = 199; // 5us x 200 = 1ms
	T2CONbits.TMR2ON = 1;

	// Timer1(RE1操作表示)
	T1CONbits.TMR1CS = 0b11; // Timerx clock Source is LFINTOSC
	T1CONbits.TMR1ON = 1;

	// Timer3(ローバッテリーLED表示点滅)
	T3CONbits.TMR3CS = 0b11; // Timerx clock Source is LFINTOSC

	// I2C(LCD)
	RB4PPS = 0b11001; // Rxy source is SDO1/SDA1
	RB6PPS = 0b11000; // Rxy source is SCK1/SCL
	SSP1ADD = 19; // 100KHz (Fosc:8MHz)
	SSP1CON1bits.SSPM = 0b1000; // I2C Master mode, clock = FOSC / (4 * (SPPxADD+1))
	SSP1CON1bits.SSPEN = 1;

	// 空きピン
	ANSELAbits.ANSA0 = 1; // pin.19
	ANSELAbits.ANSA1 = 1; // pin.18
	ANSELAbits.ANSA4 = 1; // pin.3
}

// ●システム初期化
static void InitSystem(void)
{
	// 電圧が上がるまで待つ
	SetLED(1);
	DelayMs(50);

	// AD9833の初期化
	AD9833_Reset(DDS_FRQ);
	AD9833_Reset(DDS_SIG);

	// LCD(AQM0802)の初期化
	InitLCD();

	// 初期表示
	SetLCD(0, 0, "AM Trans");
	DelayMs(1000);
	
	// 周波数設定ロード
	g_nFrq = ReadEEP(ADDR_FRQ);
	if (g_nFrq >= MAX_FRQ)
	{
		g_nFrq = DEF_FRQ;
		WriteEEP(ADDR_FRQ, (BYTE)g_nFrq);
	}
	
	g_nSig = ReadEEP(ADDR_SIG);
	if (g_nSig >= MAX_SIG)
	{
		g_nSig = DEF_SIG;
		WriteEEP(ADDR_SIG, (BYTE)g_nSig);
	}
	
	// 未発振時のAGC値を最大値として記録
	g_wMaxLv = GetADC(AGC_V);
}

// ●実行ループ
static void Run(void)
{
	BOOL bUpdate1 = TRUE;
	BOOL bUpdate2 = TRUE;

	TMR1 = SIGDISPTIME;
	TMR1IF = 0;

	int nLoop = 0;
	for(;; bUpdate1 = FALSE, bUpdate2 = FALSE, nLoop++)
	{
		// SW1入力
		BOOL fSW1 = GetSW();

		static BOOL s_fSW1 = FALSE;
		if (s_fSW1 != fSW1)
		{
			g_nSig = (fSW1)? 255 : ReadEEP(ADDR_SIG);
			s_fSW1 = fSW1;

			bUpdate2 = TRUE;
			TMR1 = SIGDISPTIME;
			TMR1IF = 0;
		}
		else
		{
			// RE入力
			switch(GetRE())
			{
			case RE1_UP:
				if (!fSW1)
				{
					g_nSig = (g_nSig < (MAX_SIG - 1))? (g_nSig + 1) : 0;
					WriteEEP(ADDR_SIG, (BYTE)g_nSig);
				}
				bUpdate2 = TRUE;
				break;
			case RE1_DOWN:
				if (!fSW1)
				{
					g_nSig = (g_nSig > 0)? (g_nSig - 1) : (MAX_SIG - 1);
					WriteEEP(ADDR_SIG, (BYTE)g_nSig);
				}
				bUpdate2 = TRUE;
				break;
			case RE2_UP:
				g_nFrq = (g_nFrq < (MAX_FRQ - 1))? (g_nFrq + 1) : 0;
				WriteEEP(ADDR_FRQ, (BYTE)g_nFrq);
				bUpdate1 = TRUE;
				break;
			case RE2_DOWN:
				g_nFrq = (g_nFrq > 0)? (g_nFrq - 1) : (MAX_FRQ - 1);
				WriteEEP(ADDR_FRQ, (BYTE)g_nFrq);
				bUpdate1 = TRUE;
				break;
			}
			if (bUpdate2)
			{
				TMR1 = SIGDISPTIME;
				TMR1IF = 0;
			}
		}
	
		// AD9833(IC3)設定 (RE2操作による送信周波数設定)
		static int s_nFrq = 255;
		if (s_nFrq != g_nFrq)
		{
			AD9833_SetFreq(DDS_FRQ, FRQ_TBL[(BYTE)g_nFrq]);
			if (AD9833[DDS_FRQ].RESET == 1) {
				AD9833_Start(DDS_FRQ);
			}
			s_nFrq = g_nFrq;
		}

		// AD9833(IC4)設定 (RE1操作による信号周波数設定)
		static int s_nSig = 255;
		if (s_nSig != g_nSig)
		{
			if (g_nSig != 255)
			{
				AD9833_SetFreq(DDS_SIG, SIG_TBL[(BYTE)g_nSig]);
				if (AD9833[DDS_SIG].RESET == 1) {
					AD9833_Start(DDS_SIG);
				}
			}
			else {
				// AUDIO入力のときはDDSをSleepにする
				AD9833_Reset(DDS_SIG);
			}
			s_nSig = g_nSig;
		}

		// 操作表示状態更新
		int nOpedisp = TMR1IF;
		static int s_nLastOpe = 0;
		if (s_nLastOpe != nOpedisp)
		{
			s_nLastOpe = nOpedisp;
			bUpdate2 = TRUE;
		}

		// 送信レベル取得
		static WORD s_wLevel = 0;

		WORD wTemp = GetADC(AGC_V);
		WORD wLevel = (wTemp < g_wMaxLv)? (g_wMaxLv - wTemp) : 0;
		wLevel = (s_wLevel + wLevel) / 2;

		if (s_wLevel != wLevel)
		{
			s_wLevel = wLevel;
			if (nOpedisp == 1) {
				if ((nLoop % 10) == 0) {
					bUpdate2 = TRUE;
				}
			}
		}

		// LCD表示更新
		if (bUpdate1)
		{
			if (g_nFrq & 1)
			{
				SetLCD(0, 0, "\x1B    .5k");
				Print(0, 4, (WORD)(513 + 9 * (g_nFrq / 2) + 4));
			}
			else
			{
				SetLCD(0, 0, "\x1B    .0k");
				Print(0, 4, (WORD)(513 + 9 * (g_nFrq / 2)));
			}
		}
		
		if (bUpdate2 )
		{
			if (nOpedisp == 0)
			{
				// 操作表示
				if (!fSW1)
				{
					if (g_nSig <= 7)
					{
						SetLCD(1, 0, "SG:  0Hz");
						Print(1, 4, (WORD)(g_nSig + 2));
					}
					else if (g_nSig <= 16)
					{
						SetLCD(1, 0, "SG: 00Hz");
						Print(1, 3, (WORD)(g_nSig - 7));
					}
					else
					{
						SetLCD(1, 0, "SG:  KHz");
						Print(1, 4, (WORD)(g_nSig - 16));
					}
				}
				else {
					SetLCD(1, 0, "   AUDIO");
				}
			}
			else
			{
				SetLCD(1, 0, "  LV:   ");
				Print(1, 7, (s_wLevel < 999)? s_wLevel : 999);
			}
		}

		// 電池残容量
		if (!T3CONbits.TMR3ON)
		{
			if (GetADC(BAT_V) < LOBAT)
			{
				T3CONbits.TMR3ON = 1;
				TMR3IF = 1;
			}
		}
		if (TMR3IF)
		{
			// ローバッテリー点滅
			SetLED(!LED);
			TMR3 = 65535 - 15500; // 約0.5秒
			TMR3IF = 0;
		}
	}
}

// ●数値をLCDに表示する
static void Print(BYTE y, BYTE r, WORD wVal)
{
	char buf[8];

	BOOL disp = FALSE;
	BYTE i = 0;
	WORD d;
	for (d = 1000; d != 0; d /= 10)
	{
		BYTE v = wVal / d;
		if (v != 0) {
			disp = TRUE;
		}
		if (disp) {
			buf[i++] = '0' + v;
			wVal = wVal % d;
		}
	}
	if (!disp) {
		buf[i++] = '0';
	}
	buf[i] = '\0';

	SetLCD(y, (BYTE)(r - i + 1), buf);
}

// ●RE入力
static int GetRE(void)
{
	int nRet = GetRE1();
	return (nRet != RE_NONE)? nRet : GetRE2();
}

// ●RE1入力
static int GetRE1(void)
{
	static BOOL s_bAsig1 = FALSE;

	// A signal が立ち下がった時の B signal を調べる
	BOOL bCurA = RE1_A;
	if (bCurA != s_bAsig1)
	{
		s_bAsig1 = bCurA;

		if (!bCurA) {
			return (RE1_B)? RE1_UP : RE1_DOWN;
		}
	}

	return RE_NONE;
}

// ●RE1入力
static int GetRE2(void)
{
	static BOOL s_bAsig1 = FALSE;

	// A signal が立ち下がった時の B signal を調べる
	BOOL bCurA = RE2_A;
	if (bCurA != s_bAsig1)
	{
		s_bAsig1 = bCurA;

		if (!bCurA) {
			return (RE2_B)? RE2_UP : RE2_DOWN;
		}
	}

	return RE_NONE;
}

// ●SW入力
static BOOL GetSW(void)
{
	return SW1;
}

// ●ADC値取得
static WORD GetADC(BYTE nCh)
{
	ADCON0bits.CHS = nCh;
	
	// Wait 10uS
	NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP();
	NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP();

	ADCON0bits.GO = 1;
	while(ADCON0bits.GO_nDONE);
	
	return ADRES;
}

// ●LED設定
static void SetLED(int nVal)
{
	LED = nVal;
}

// ●AD9833初期化
static void AD9833_Reset(BYTE nNo)
{
	AD9833reg_t* pAD9833 = &AD9833[nNo];
	
	pAD9833->wValue = 0;

	pAD9833->B28 = 1; // 常に28bit書き込み
	pAD9833->RESET = 1;
	pAD9833->SLEEP1 = 1; // Internal clock disabled
	pAD9833->SLEEP12 = 1; // DAC powered down
	
	AD9833_Send(nNo, pAD9833->wValue);
}

// ●AD9833周波数設定
static void AD9833_SetFreq(BYTE nNo, DWORD dwVal)
{
	AD9833reg_t* pAD9833 = &AD9833[nNo];
	AD9833reg_t Reg1, Reg2;

	Reg1.wValue = dwVal & 0x3FFF;
	Reg2.wValue = dwVal >> 14;
	
	if (pAD9833->FSEL == 0)
	{
		// FREQ1へ書き込む
		pAD9833->FSEL = 1;
		Reg1.D15 = 1;
		Reg2.D15 = 1;
	}
	else
	{
		// FREQ0へ書き込む
		pAD9833->FSEL = 0;
		Reg1.D14 = 1;
		Reg2.D14 = 1;
	}

	AD9833_Send(nNo, Reg1.wValue);
	AD9833_Send(nNo, Reg2.wValue);

	// FREQ 0/1切り替え
	AD9833_Send(nNo, pAD9833->wValue);	
}

// ●AD9833動作開始
static void AD9833_Start(BYTE nNo)
{
	AD9833reg_t* pAD9833 = &AD9833[nNo];

	pAD9833->RESET = 0;
	pAD9833->SLEEP1 = 0;
	pAD9833->SLEEP12 = 0;
	
	AD9833_Send(nNo, pAD9833->wValue);	
}

// ●AD9833への送信
static void AD9833_Send(BYTE nNo, WORD wData)
{
	if (nNo == DDS_FRQ) AD9833_SYNC1 = 0;
	else AD9833_SYNC2 = 0;

	int nCount = 0;
	do
	{
		AD9833_SDAT = (wData & 0x8000)? 1 : 0;
		DelayNs(5); // t7, t9
		
		AD9833_SCLK = 0;
		DelayNs(10); // t6
		AD9833_SCLK = 1;
		DelayNs(5); // t5-t9
		
		wData <<= 1;
	}
	while (++nCount < 16);
	
	if (nNo == DDS_FRQ) AD9833_SYNC1 = 1;
	else AD9833_SYNC2 = 1;
	
	DelayNs(5); // t11	
}

// ●LCD(AQM0802)を初期化する
static void InitLCD(void)
{
	CmdLCD(0x38);
	CmdLCD(0x39);
	CmdLCD(0x14);
	CmdLCD(0x70);
	CmdLCD(0x56);
	CmdLCD(0x6C);
	DelayMs(200);
	CmdLCD(0x38);
	CmdLCD(0x0C);
	CmdLCD(0x01);
	DelayMs(2);

	SetLCD(0, 0, "        ");
	SetLCD(2, 0, "        ");
}

// ●LCD(AQM0802)へ文字列を出力する
static void SetLCD(BYTE y, BYTE x, const char* pDat)
{
	I2CStart();

	I2CSend(ADDR_LCD << 1);
	I2CSend(0b10000000); // Co:1, RS:0
	I2CSend((BYTE)(0b10000000 + y * 0x40 + x)); // Set DDRAM address

	I2CSend(0b01000000); // Co:0, RS:1

	const char* p;
	for (p = pDat; *p != '\0'; p++) {
		I2CSend(*p);
	}
	
	I2CStop();
}

// ●LCD(AQM0802)へコマンドを送信する
static void CmdLCD(BYTE bCmd)
{
	I2CStart();

	I2CSend(ADDR_LCD << 1);
	I2CSend(0b10000000); // Co:1, RS:0
	I2CSend(bCmd);

	I2CStop();
}

// ●I2Cスタート条件
static void I2CStart(void)
{
	SSP1IF = 0;
    SSP1CON2bits.SEN = 1;
    while (!SSP1IF);	
}

// ●I2C送信
static void I2CSend(BYTE bData)
{
	SSP1IF = 0;
	SSP1BUF = bData;
	while (!SSP1IF);
}

// ●I2Cストップ条件
static void I2CStop(void)
{
	SSP1IF = 0;
    SSP1CON2bits.PEN = 1;
    while (!SSP1IF);	
}

// ●EEPROMへ書き出し
static void WriteEEP(BYTE nAdd, BYTE nData)
{
	NVMCON1bits.NVMREGS = 1;
	NVMADR = 0x7000UL + nAdd;
	
	NVMCON1bits.FREE = 0;
	NVMCON1bits.LWLO = 0;
	NVMCON1bits.WREN = 1;
	
	NVMDATL = nData;
	
	NVMCON2 = 0x55;
	NVMCON2 = 0xAA;
	NVMCON1bits.WR = 1;
    while(NVMCON1bits.WR);
	
	NVMCON1bits.WREN = 0;
}

// ●EEPROMから読み出し
static BYTE ReadEEP(BYTE nAdd)
{
	NVMCON1bits.NVMREGS = 1;
	NVMADR = 0x7000UL + nAdd;
	
	NVMCON1bits.RD = 1;
	NOP();
	
	return NVMDATL;
}

// ●MS待ち
static void DelayMs(WORD wMs)
{
	TMR2 = 0;
	do {
		TMR2IF = 0;
		while (!TMR2IF);
	}
	while (--wMs > 0);
}
